﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using System.Text.RegularExpressions;
using System.Security;
using System.Reflection;
/* PowerShell */
using System.Management.Automation;
using System.Management.Automation.Runspaces;
/* For IP related tasks */
using System.Net;
using System.Net.Sockets;
/* Powershell AD modules may be missing in some environments so use Directory Services */
using System.DirectoryServices.AccountManagement;

namespace viperPS
{
    class PSWork
    {
        /// <summary>
        /// Get the machine name (aka HostName).  Accepts HostName and returns that provided HostName.  Accepts IPv4 or IPv6 address and converts to HostName.  Accepts nothing and assumes the local computer.  Accepts ::1 and assumes the local computer.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        private static string PSGetComputerName(string computerName = "")
        {
            string tmpComputerName = String.Empty;
            string tmpUserTargetComputer = computerName.ToString().Trim();
            if (tmpUserTargetComputer.Length > 0 && IsValidComputerNameIPString(tmpUserTargetComputer) == false && tmpUserTargetComputer != "$env:ComputerName")
            {
                /* Computer name not valid */
                tmpComputerName = "[ERROR] Computer name or IP appears to be invalid.";
            }
            else if (tmpUserTargetComputer == "::1" || tmpUserTargetComputer == "127.0.0.1" || tmpUserTargetComputer.Length == 0)
            {
                /* Local computer syntax or nothing provided */
                tmpComputerName = "$env:ComputerName";
            }
            else if (IsValidIP(tmpUserTargetComputer) == true)
            {
                /* IPv4 or IPv6 address entered for computer name.  Get computer name. */
                string psErrorData = String.Empty;
                try
                {
                    StringBuilder psResult = new StringBuilder();
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript(String.Format("[System.Net.Dns]::GetHostEntry(\"{0}\").Hostname", tmpUserTargetComputer));
                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.Append(obj.ToString());
                    }
                    tmpComputerName = psResult.ToString();

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = "[ERROR](Target Computer Invalid) " + ex.Message.ToString();
                    }
                }
                if (psErrorData.Length > 0)
                {
                    tmpComputerName = psErrorData;
                }
            }
            else if (IsValidIP(tmpUserTargetComputer) == false)
            {
                if (tmpUserTargetComputer.IndexOf(".", StringComparison.OrdinalIgnoreCase) > -1)
                {
                    /* Computer name entered may have the dns suffix attached to it, if on a domain.  Drop the suffix. */
                    tmpUserTargetComputer = tmpUserTargetComputer.Split('.')[0];
                }

                /* Try the resultant, assumed, computer name.  Verify the computer name by retrieving IP */
                string psErrorData = String.Empty;
                try
                {
                    StringBuilder psResult = new StringBuilder();
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript(String.Format("[System.Net.Dns]::GetHostAddresses(\"{0}\").IPAddressToString", tmpUserTargetComputer));
                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    bool tmpAddressReturned = false;
                    foreach (PSObject obj in results)
                    {
                        if (IsValidIP(obj.ToString()) == true) { tmpAddressReturned = true; }
                    }
                    if (tmpAddressReturned == false)
                    {
                        psErrorData = "[ERROR](Target Computer Not Found)";
                    }
                    else
                    {
                        tmpComputerName = tmpUserTargetComputer;
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = "[ERROR](Target Computer Not Found) " + ex.Message.ToString();
                    }
                }
                if (psErrorData.Length > 0)
                {
                    tmpComputerName = psErrorData;
                }
            }
            else
            {
                /* The submitted computer name or IP does not appear to be valid for use */
                tmpComputerName = "[ERROR](Target Computer Not Found) Computer name or ip appears to be invalid";
            }
            return tmpComputerName;
        }

        /// <summary>
        /// Get the logged in user of the target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetLoggedInUser(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript(String.Format("(Get-WmiObject -Class win32_computersystem -ComputerName \"{0}\").Username", tmpComputerName));
                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.Append(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Gets the SID of user logged into target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetUserSID(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript(String.Format("([System.Security.Principal.NTAccount]((Get-WmiObject -Class win32_computersystem -ComputerName \"{0}\").UserName)).Translate([System.Security.Principal.SecurityIdentifier]).Value", tmpComputerName));
                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.Append(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Logs off users from target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSLogOffUser(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    if (tmpComputerName.ToLower().Trim() == "$env:computername")
                    {
                        powershell.AddScript("Invoke-Command -ScriptBlock { shutdown.exe /l }");
                    }
                    else
                    {
                        powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\"", tmpComputerName) + " -ScriptBlock { shutdown.exe /l }");
                    }

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.Append(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Provide response message applicable in this case */
            if (psErrorData.Length == 0)
            {
                psResult.Clear();
                psResult.Append("Log off completed.");
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Restarts target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSRestartMachine(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript(String.Format("Restart-Computer -ComputerName \"{0}\" -Force", tmpComputerName));

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.Append(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Provide response message applicable in this case */
            if (psErrorData.Length == 0)
            {
                psResult.Clear();
                psResult.Append("Target Machine Restarted.");
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Locks target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSLockMachine(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript(String.Format("Invoke-WmiMethod -ComputerName \"{0}\" -Class Win32_Process -Name Create -ArgumentList \"C:\\Windows\\System32\\rundll32.exe user32.dll,LockWorkStation\"", tmpComputerName));

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.Append(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Provide response message applicable in this case */
            if (psErrorData.Length == 0)
            {
                psResult.Clear();
                psResult.Append("Target Machine Locked.");
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Gets disk space information of target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetDiskspaceMachine(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript("Get-WmiObject Win32_LogicalDisk -ComputerName " + String.Format("\"{0}\"", tmpComputerName) + " -Filter DriveType=3 | Select-Object DeviceID, @{'Name'='Size (GB)'; 'Expression'={[string]::Format('{0:N0}',[math]::truncate($_.size / 1GB))}}, @{'Name'='Freespace (GB)'; 'Expression'={[string]::Format('{0:N0}',[math]::truncate($_.freespace / 1GB))}}");

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.AppendLine(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Gets name of target machine when supplied with an IP address (v4 or v6)
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetMachineName(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            if (IsValidIP(computerName.ToString().Trim()) == true)
            {
                /* Get computer name from IP address */
                tmpComputerName = PSGetComputerName(computerName);
            }
            else
            {
                /* Get computer name of local computer since nothing was provided */
                tmpComputerName = PSGetComputerName(computerName);
            }

            /* Do not continue if an error was found */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                if (tmpComputerName == "$env:ComputerName")
                {
                    /* No IP address was provided so get the local computer name */
                    try
                    {
                        var runspace = RunspaceFactory.CreateRunspace();
                        runspace.Open();
                        var powershell = PowerShell.Create();
                        powershell.Runspace = runspace;

                        powershell.AddScript("$env:ComputerName");

                        var results = powershell.Invoke();
                        /* Catch remote errors such as no permissions */
                        foreach (ErrorRecord current in powershell.Streams.Error)
                        {
                            psErrorData = psErrorData + current.Exception.ToString();
                        }
                        if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                        /* Gather response data */
                        foreach (PSObject obj in results)
                        {
                            psResult.Append(obj.ToString());
                        }

                        powershell.Dispose();
                        runspace.Dispose();
                    }
                    catch (Exception ex)
                    {
                        if (ex.Message.Length > 0)
                        {
                            psErrorData = ex.Message.ToString();
                        }
                    }
                }
                else
                {
                    psResult.Append(tmpComputerName);
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Gets the IP address(s) of the target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetMachineIP(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
             tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    if (tmpComputerName == "$env:ComputerName")
                    {
                        powershell.AddScript(String.Format("[System.Net.Dns]::GetHostAddresses({0}).IPAddressToString", tmpComputerName));
                    }
                    else
                    {
                        powershell.AddScript(String.Format("[System.Net.Dns]::GetHostAddresses(\"{0}\").IPAddressToString", tmpComputerName));
                    }                    

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.AppendLine(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Gets hardware and information regarding the target computer.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetMachineInfo(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);

            /* Get IP address(s) of target computer */
            StringBuilder psIPResult = new StringBuilder();
            string responseData = String.Empty;
            string responseError = String.Empty;
            List<KeyValuePair<string, string>> actionResult = PSGetMachineIP(tmpComputerName);
            foreach (KeyValuePair<string, string> keyValuePairs in actionResult)
            {
                if (keyValuePairs.Key.ToLower() == "psdata") { responseData = keyValuePairs.Value; }
                if (keyValuePairs.Key.ToLower() == "error") { responseError = keyValuePairs.Value; }
            }
            if (responseError.Length > 0)
            {
                psIPResult.Append(responseError);
            }
            else
            {
                psIPResult.Append(responseData);
            }

            /* Get MAC address(s) of target computer */
            StringBuilder psMacResult = new StringBuilder();
            psMacResult = GetMacAddress(tmpComputerName);

            /* Get Serial number of target computer */
            StringBuilder psBiosResult = new StringBuilder();
            psBiosResult = GetBiosInfo(tmpComputerName);

            /* Get Hardware information of target computer */
            StringBuilder psHardwareResult = new StringBuilder();
            psHardwareResult = GetHardwareInfo(tmpComputerName);

            /* Get Processor information of target computer */
            StringBuilder psProcessorResult = new StringBuilder();
            psProcessorResult = GetProcessorInfo(tmpComputerName);

            /* Get OS information of target computer */
            StringBuilder psOSResult = new StringBuilder();
            psOSResult = GetOSInfo(tmpComputerName);

            /* Get OU information of target computer */
            StringBuilder psOUResult = new StringBuilder();
            psOUResult = GetOUInfo(tmpComputerName);

            /* Compile results */
            psResult.AppendLine("IP ADDRESS:");
            psResult.Append(psIPResult);
            psResult.AppendLine();

            psResult.AppendLine("MAC ADDRESS:");
            psResult.Append(psMacResult);
            psResult.AppendLine();

            psResult.AppendLine("SERIAL NUMBER:");
            psResult.Append(psBiosResult);
            psResult.AppendLine();

            psResult.AppendLine("HARDWARE:");
            psResult.Append(psHardwareResult);
            psResult.AppendLine();

            psResult.AppendLine("PROCESSOR:");
            psResult.Append(psProcessorResult);
            psResult.AppendLine();

            psResult.AppendLine("ORGANIZATIONAL UNIT:");
            psResult.Append(psOUResult);

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Get uptime of target computer.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetMachineUptime(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript(String.Format("(Get-Date) - [Management.ManagementDateTimeConverter]::ToDateTime((Get-WmiObject Win32_OperatingSystem -ComputerName \"{0}\").LastBootUpTime) | Select-Object Ticks,TotalDays,TotalHours,TotalMinutes,TotalSeconds,TotalMilliseconds", tmpComputerName));

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.AppendLine(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Get processes of target machine and target user.
        /// </summary>
        /// <param name="computerName"></param>
        /// <param name="userName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetMachineProcesses(string computerName = "", string userName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            string tmpUserName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Get username */
            tmpUserName = userName.Trim();
                /* Do not continue if the computer name could not be resolved */
                if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
                {
                    try
                    {
                        var runspace = RunspaceFactory.CreateRunspace();
                        runspace.Open();
                        var powershell = PowerShell.Create();
                        powershell.Runspace = runspace;

                        if (tmpUserName.Length > 0 && IsValidString(tmpUserName) == true)
                        {
                            powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\"", tmpComputerName) + " -ScriptBlock { Get-Process -IncludeUserName | Where-Object {$_.username match \"" + String.Format("{0}", tmpUserName) + "\"} }");
                        }
                        else
                        {
                            powershell.AddScript("Get-Process | Select-Object Id,SI,ProcessName");
                        }

                        var results = powershell.Invoke();
                        /* Catch remote errors such as no permissions */
                        foreach (ErrorRecord current in powershell.Streams.Error)
                        {
                            psErrorData = psErrorData + current.Exception.ToString();
                        }
                        if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                        /* Gather response data */
                        foreach (PSObject obj in results)
                        {
                            psResult.AppendLine(obj.ToString());
                        }
                        if (psErrorData.Length == 0 && tmpUserName.Length > 0 && IsValidString(tmpUserName) == false)
                        {
                        psErrorData = "[ERROR] Target username appears to be invalid so the current username on the local machine was used.";
                        }

                        powershell.Dispose();
                        runspace.Dispose();
                    }
                    catch (Exception ex)
                    {
                        if (ex.Message.Length > 0)
                        {
                            psErrorData = ex.Message.ToString();
                        }
                    }
                }
                else
                {
                    psErrorData = tmpComputerName;
                }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Start RDP session to target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSRemoteToMachine(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript(String.Format("Start-Process \"$env:windir\\system32\\mstsc.exe\" -ArgumentList \"/v:{0}\"", tmpComputerName));

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.AppendLine(obj.ToString());
                    }
                    /* If no errors were found, assume mstsc launched successfully */
                    if (psErrorData.Length == 0)
                    {
                        psResult.Clear();
                        psResult.Append("Request Completed.");
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Starts Remote Assistance session to target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSRemoteAssistanceToMachine(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript(String.Format("Start-Process msra.exe -ArgumentList \"/OfferRA {0}\"", tmpComputerName));

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.AppendLine(obj.ToString());
                    }
                    /* If no errors were found, assume mstsc launched successfully */
                    if (psErrorData.Length == 0)
                    {
                        psResult.Clear();
                        psResult.Append("Request Completed.");
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Performs ICMP Ping to target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSPingMachine(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    powershell.AddScript(String.Format("Test-Connection -ComputerName \"{0}\"", tmpComputerName) + " -Count 1 | Select-Object IPV4Address,IPV6Address,StatusCode,@{'Name'='Status Detail'; 'Expression'={switch($_.StatusCode){ 0{\"OK\";break} 1101{\"Buffer Too Small\";break} 11002{\"Destination Net Unreachable\";break} 11003{\"Destination Host Unreachable\";break} 11004{\"Destination Protocol Unreachable\";break} 11005{\"Destination Port Unreachable\";break} 11006{\"No Resources\";break} 11007{\"Bad Option\";break} 11008{\"Hardware Error\";break} 11009{\"Packet Too Big\";break} 11010{\"Request Timed Out\";break} 11011{\"Bad Request\";break} 11012{\"Bad Route\";break} 11013{\"TimeToLive Expired Transit\";break} 11014{\"TimeToLive Expired Reassembly\";break} 11015{\"Parameter Problem\";break} 11016{\"Source Quench\";break} 11017{\"Option Too Big\";break} 11018{\"Bad Destination\";break} 11032{\"Negotiating IPSEC\";break} 11050{\"General Failure\";break} }}}");

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.AppendLine(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Performs Trace route to target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSTraceRouteMachine(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;
                    /* Determine whether to target local machine or remote machine */
                    if (tmpComputerName == "$env:ComputerName")
                    {
                        tmpComputerName = "::1";
                    }
                    powershell.AddScript(String.Format("tracert {0}", tmpComputerName));

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.AppendLine(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Gets a specific log (Application, System, Security) from the target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <param name="logName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetMachineEventLogs(string computerName = "", string logName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            string tmpLogName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Determine logName to use */
            switch(logName)
            {
                case "Application":
                    tmpLogName = "Application";
                    break;
                case "Security":
                    tmpLogName = "Security";
                    break;
                case "System":
                    tmpLogName = "System";
                    break;
                default:
                    tmpLogName = "";
                    break;
            }
            /* Do not continue if the computer name could not be resolved or if a log type was not specified */
            if (tmpLogName.Length > 0)
            {
                if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
                {
                    try
                    {
                        var runspace = RunspaceFactory.CreateRunspace();
                        runspace.Open();
                        var powershell = PowerShell.Create();
                        powershell.Runspace = runspace;

                        /*
                         * NOTE: Below could also be used via command-line.  However, retrieval of the Security log still requires additional permissions and without those permissions access will be denied.
                         * powershell.AddScript("Invoke-Command -ScriptBlock { Start-Process wevtutil qe Security /c:100 /rd:true /f:text }");
                         */
                        powershell.AddScript(String.Format("Get-EventLog -ComputerName \"{0}\" -LogName \"{1}\" -EntryType Error -Newest 100 | Select-Object Time,EntryType,Source,InstanceID,Message", tmpComputerName, tmpLogName));

                        var results = powershell.Invoke();
                        /* Catch remote errors such as no permissions */
                        foreach (ErrorRecord current in powershell.Streams.Error)
                        {
                            psErrorData = psErrorData + current.Exception.ToString();
                        }
                        if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                        /* Gather response data */
                        foreach (PSObject obj in results)
                        {
                            psResult.AppendLine(obj.ToString());
                        }

                        powershell.Dispose();
                        runspace.Dispose();
                    }
                    catch (Exception ex)
                    {
                        if (ex.Message.Length > 0)
                        {
                            psErrorData = ex.Message.ToString();
                        }
                    }
                }
                else
                {
                    psErrorData = tmpComputerName;
                }
            }
            else
            {
                /* Log type not specified */
                psErrorData = "[ERROR] Log type must be specified.";
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Runs a bat or ps1 file on the target computer.
        /// </summary>
        /// <param name="computerName"></param>
        /// <param name="localFile"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSRunBatPsTargetMachine(string computerName = "", string localFile = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            string tmpLocalFile = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Validate the localFile and ensure it is a bat or ps1 file */
            bool validatedPath = false;
            validatedPath = IsValidFilepath(localFile);
            string extension = String.Empty;
            if (validatedPath == true)
            {
                bool eValid = false;
                string[] segments = localFile.Split('.');
                extension = segments[segments.Length - 1].ToString().ToLower();
                switch (extension)
                {
                    case "bat":
                        eValid = true;
                        break;
                    case "ps1":
                        eValid = true;
                        break;
                    default:
                        eValid = false;
                        break;
                }
                if (eValid == true)
                {
                    tmpLocalFile = localFile;
                }
            }

            /* Do not continue if the local file was not valid */
            if (tmpLocalFile.Length > 0)
            {
                if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
                {
                    try
                    {
                        var runspace = RunspaceFactory.CreateRunspace();
                        runspace.Open();
                        var powershell = PowerShell.Create();
                        powershell.Runspace = runspace;

                        /* Determine if script is to be run locally or on target machine */
                        if (tmpComputerName == "$env:ComputerName")
                        {
                            if (extension == "bat")
                            {
                                powershell.AddScript("Invoke-Command -ScriptBlock { Start-Process " + String.Format("'{0}'", tmpLocalFile) + " }");
                            }
                            else
                            {
                                powershell.AddScript("Invoke-Command -ScriptBlock { Start-Process PowerShell " + String.Format("\"{0}\"", tmpLocalFile) + " }");
                            }
                        }
                        else
                        {
                            if (extension == "bat")
                            {
                                powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\" -ScriptBlock ", tmpComputerName) + "{ Start-Process " + String.Format("'{0}'", tmpLocalFile) + " }");
                            }
                            else
                            {
                                powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\" -ScriptBlock ", tmpComputerName) + "{ Start-Process PowerShell " + String.Format("\"{0}\"", tmpLocalFile) + " }");
                            }
                        }
                        
                        var results = powershell.Invoke();
                        /* Catch remote errors such as no permissions */
                        foreach (ErrorRecord current in powershell.Streams.Error)
                        {
                            psErrorData = psErrorData + current.Exception.ToString();
                        }
                        if (psErrorData.Length == 0)
                        {
                            psResult.Clear();
                            psResult.Append("Request completed in separate window which will stay open only if your script contains a pause or waits for manual interaction.");
                        }

                        powershell.Dispose();
                        runspace.Dispose();
                    }
                    catch (Exception ex)
                    {
                        if (ex.Message.Length > 0)
                        {
                            psErrorData = ex.Message.ToString();
                        }
                    }
                }
                else
                {
                    psErrorData = tmpComputerName;
                }
            }
            else
            {
                /* File was not valid */
                psErrorData = "[ERROR] File specified was not valid.";
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Runs a bat or ps1 file on the target machinens specified by the computer list file.
        /// </summary>
        /// <param name="computerList"></param>
        /// <param name="localFile"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSRunBatPsTargetMachines(string computerList = "", string localFile = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerList = String.Empty;
            string tmpLocalFile = String.Empty;

            /* Validate the computerList and ensure it is a txt file */
            bool validatedList = false;
            validatedList = IsValidFilepath(computerList);
            string extensionList = String.Empty;
            if (validatedList == true)
            {
                bool eValid = false;
                string[] segments = computerList.Split('.');
                extensionList = segments[segments.Length - 1].ToString().ToLower();
                switch (extensionList)
                {
                    case "txt":
                        eValid = true;
                        break;
                    default:
                        eValid = false;
                        break;
                }
                if (eValid == true)
                {
                    tmpComputerList = computerList;
                }

            }
            
            /* Create path to FailedComputers.txt (in the same location as the computer list file) in the event any computers fail to run the script */
            string tmpComputerListFailed = String.Empty;
            if (tmpComputerList.Length > 0)
            {
                var lastBackslash = tmpComputerList.LastIndexOf('\\');
                tmpComputerListFailed = tmpComputerList.Substring(0, lastBackslash);
                tmpComputerListFailed = tmpComputerListFailed + "\\FailedComputers.txt";
            }

            /* Validate the localFile and ensure it is a bat or ps1 file */
            bool validatedPath = false;
            validatedPath = IsValidFilepath(localFile);
            string extension = String.Empty;
            if (validatedPath == true)
            {
                bool eValid = false;
                string[] segments = localFile.Split('.');
                extension = segments[segments.Length - 1].ToString().ToLower();
                switch (extension)
                {
                    case "bat":
                        eValid = true;
                        break;
                    case "ps1":
                        eValid = true;
                        break;
                    default:
                        eValid = false;
                        break;
                }
                if (eValid == true)
                {
                    tmpLocalFile = localFile;
                }
            }
                                          
            /* Do not continue if the local script file was not valid */
            if (tmpLocalFile.Length > 0)
            {
                /* Do not continue if the computer list file was not valid */
                if (tmpComputerList.Length > 0)
                {
                    try
                    {
                        var runspace = RunspaceFactory.CreateRunspace();
                        runspace.Open();
                        var powershell = PowerShell.Create();
                        powershell.Runspace = runspace;

                        if (extension == "bat")
                        {
                            powershell.AddScript(String.Format("ForEach ($c in (Get-Content \"{0}\")) ", tmpComputerList) + "{ Try { Invoke-Command -ComputerName $c -ScriptBlock { Start-Process " + String.Format("'{0}' -NoNewWindow -Wait", tmpLocalFile) + "} -ErrorAction Stop } Catch { Add-Content " + String.Format("\"{0}\"", tmpComputerListFailed) + " $c } }");
                        }
                        else
                        {
                            powershell.AddScript(String.Format("ForEach ($c in (Get-Content \"{0}\")) ", tmpComputerList) + "{ Try { Invoke-Command -ComputerName $c -ScriptBlock { Start-Process PowerShell " + String.Format("'{0}' -NoNewWindow -Wait", tmpLocalFile) + "} -ErrorAction Stop } Catch { Add-Content " + String.Format("\"{0}\"", tmpComputerListFailed) + " $c } }");
                        }

                        var results = powershell.Invoke();
                        /* Catch remote errors such as no permissions */
                        foreach (ErrorRecord current in powershell.Streams.Error)
                        {
                            psErrorData = psErrorData + current.Exception.ToString();
                        }
                        if (psErrorData.Length > 0) { psErrorData = "[ERROR] " + psErrorData; }
                        /* If no errors were found, assume start-process launched successfully on target machines.  Provide list of target machines that failed, if applicable. */
                        if (psErrorData.Length == 0)
                        {
                            psResult.Clear();
                            psResult.Append("Request Completed.  Because of potential resource load, an individual window to each machine was not loaded.  If the script failed on any machines, a list of failed machines will be at " + tmpComputerListFailed);
                        }

                        powershell.Dispose();
                        runspace.Dispose();
                    }
                    catch (Exception ex)
                    {
                        if (ex.Message.Length > 0)
                        {
                            psErrorData = ex.Message.ToString();
                        }
                    }
                }
                else
                {
                    /* Computer list file was not valid. */
                    psErrorData = "[ERROR] File specified for computer list was not valid.";
                }
            }
            else
            {
                /* File was not valid */
                psErrorData = "[ERROR] File specified for script file was not valid.";
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Opens compmgt.msc on the target computer.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSMachineManagement(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;
                    if (tmpComputerName == "$env:ComputerName")
                    {
                        powershell.AddScript("Invoke-Command -ScriptBlock { Start-Process \"compmgmt.msc\" }");
                    }
                    else
                    {
                        powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\"", tmpComputerName) + " -ScriptBlock { Start-Process \"compmgmt.msc\" }");
                    }

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    if (psErrorData.Length == 0)
                    {
                        psResult.Append("Request Completed.");
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Opens a command-line or powershell window on the target machine for the user to manually enter commands.
        /// </summary>
        /// <param name="computerName"></param>
        /// <param name="runRemoteCommand"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSRunRemoteCommand(string computerName = "", string runRemoteCommand = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            string tmpRunRemoteCommand = String.Empty;

            /* User computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Determine shell enviornment to use */
            switch (runRemoteCommand)
            {
                case "Command Line - CMD":
                    tmpRunRemoteCommand = "cmd";
                    break;
                case "PowerShell - PS":
                    tmpRunRemoteCommand = "ps";
                    break;
                default:
                    tmpRunRemoteCommand = "";
                    break;
            }

            /* Do not continue if environment was not selected or if computer name could not be resolved */
            if (tmpRunRemoteCommand.Length > 0)
            {
                if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
                {
                    try
                    {
                        var runspace = RunspaceFactory.CreateRunspace();
                        runspace.Open();
                        var powershell = PowerShell.Create();
                        powershell.Runspace = runspace;

                        if (tmpComputerName == "$env:ComputerName")
                        {
                            if (tmpRunRemoteCommand == "cmd")
                            {
                                powershell.AddScript("Invoke-Command -ScriptBlock { Start-Process \"cmd.exe\" }");
                            }
                            else
                            {
                                powershell.AddScript("Invoke-Command -ScriptBlock { Start-Process PowerShell }");
                            }
                        }
                        else
                        {
                            if (tmpRunRemoteCommand == "cmd")
                            {
                                powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\"", tmpComputerName) + " -ScriptBlock { Start-Process \"cmd.exe\" }");
                            }
                            else
                            {
                                powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\"", tmpComputerName) + " -ScriptBlock { Start-Process PowerShell }");
                            }
                        }
                        
                        var results = powershell.Invoke();
                        /* Catch remote errors such as no permissions */
                        foreach (ErrorRecord current in powershell.Streams.Error)
                        {
                            psErrorData = psErrorData + current.Exception.ToString();
                        }
                        if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                        /* Gather response data */
                        if (psErrorData.Length == 0)
                        {
                            psResult.Append("Request Completed");
                        }

                        powershell.Dispose();
                        runspace.Dispose();
                    }
                    catch (Exception ex)
                    {
                        if (ex.Message.Length > 0)
                        {
                            psErrorData = ex.Message.ToString();
                        }
                    }
                }
                else
                {
                    psErrorData = tmpComputerName;
                }
            }
            else
            {
                /* Environment type not specified. */
                psErrorData = "[ERROR] Environment type must be specified.";
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// On Windows Pro and higher (non-home edition) computers, clear the CSC (offline files) cache on the target machine.  A reboot will be required and syncs re-established.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSCscOfflineFilesClean(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;
                    if (tmpComputerName == "$env:ComputerName")
                    {
                        powershell.AddScript("Invoke-Command -ScriptBlock { Invoke-CimMethod -ClassName Win32_OfflineFilesCache -MethodName Enable -Arguments @{Enable=$false} -ErrorAction Stop; if (-Not(Test-Path -Path ((Get-Location | Select-Object | %{$_.ProviderPath}) + \"tmpCSCFolder\") -PathType Container)) { mkdir \"tmpCSCFolder\" } robocopy ((Get-Location | Select-Object | %{$_.ProviderPath}) + \"tmpCSCFolder\") \"C:\\Windows\\CSC\\v2.0.6\\cache\" /mir; Remove-Item -LiteralPath ((Get-Location | Select-Object | %{$_.ProviderPath}) + \"tmpCSCFolder\") -Force -ErrorAction Stop; reg add HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\Csc\\Parameters /v FormatDatabase /t REG_DWORD /d 1 /f; Invoke-CimMethod -ClassName Win32_OfflineFilesCache -MethodName Enable -Arguments @{Enable=$true} -ErrorAction Stop }");
                    }
                    else
                    {
                        powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\"", tmpComputerName) + " -ScriptBlock { Invoke-CimMethod -ClassName Win32_OfflineFilesCache -MethodName Enable -Arguments @{Enable=$false} -ErrorAction Stop; if (-Not(Test-Path -Path ((Get-Location | Select-Object | %{$_.ProviderPath}) + \"tmpCSCFolder\") -PathType Container)) { mkdir \"tmpCSCFolder\" } robocopy ((Get-Location | Select-Object | %{$_.ProviderPath}) + \"tmpCSCFolder\") \"C:\\Windows\\CSC\\v2.0.6\\cache\" /mir; Remove-Item -LiteralPath ((Get-Location | Select-Object | %{$_.ProviderPath}) + \"tmpCSCFolder\") -Force -ErrorAction Stop; reg add HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\Csc\\Parameters /v FormatDatabase /t REG_DWORD /d 1 /f; Invoke-CimMethod -ClassName Win32_OfflineFilesCache -MethodName Enable -Arguments @{Enable=$true} -ErrorAction Stop }");
                    }

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    if (psErrorData.Length == 0)
                    {
                        psResult.Append("Request Completed.  The target computer should be restarted for changes to take effect.");
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Get list of services and state from target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetServices(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;
                    if (tmpComputerName == "$env:ComputerName")
                    {
                        powershell.AddScript("Get-Service | Select-Object Status,Name,DisplayName");
                    }
                    else
                    {
                        powershell.AddScript(String.Format("Get-Service -ComputerName \"{0}\" | Select-Object Status,Name,DisplayName", tmpComputerName));
                    }

                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.AppendLine(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        /// <summary>
        /// Restart service on target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <param name="targetService"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSRestartService(string computerName = "", string targetService = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            string tmpTargetService = String.Empty;

            /* Use computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Validate service name */
            if (IsValidString(targetService) == true)
            {
                tmpTargetService = targetService;
            }

            /* Do not continue if the computer name could not be resolved or service name was invalid */
            if (tmpTargetService.Length > 0)
            {
                if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
                {
                    try
                    {
                        var runspace = RunspaceFactory.CreateRunspace();
                        runspace.Open();
                        var powershell = PowerShell.Create();
                        powershell.Runspace = runspace;

                        /*
                         * NOTE: Stopping and starting a service could also be done by command-line.  However, that requires additional permissions aside from needing to remote to the target machine.
                         * powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\"", tmpComputerName) + " -ScriptBlock { " + String.Format("net stop \"{0}\"; net start \"{0}\";", tmpTargetService) + " }");
                         */
                        if (tmpComputerName == "$env:ComputerName")
                        {
                            powershell.AddScript(String.Format("Stop-Service -Name \"{0}\"; (Get-Service -Name \"{0}\").WaitForStatus('Stopped', '00:00:30'); Start-Service -Name \"{0}\"; (Get-Service -Name \"{0}\").WaitForStatus('Running', '00:00:30')", tmpTargetService));
                        }
                        else
                        {
                            powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\"", tmpComputerName) + " -ScriptBlock { " + String.Format("Stop-Service -Name \"{0}\"; (Get-Service -Name \"{0}\").WaitForStatus('Stopped', '00:00:30'); Start-Service -Name \"{0}\"; (Get-Service -Name \"{0}\").WaitForStatus('Running', '00:00:30')", tmpTargetService) + " }");
                        }

                        var results = powershell.Invoke();
                        /* Catch remote errors such as no permissions */
                        foreach (ErrorRecord current in powershell.Streams.Error)
                        {
                            psErrorData = psErrorData + current.Exception.ToString();
                        }
                        if (psErrorData.Length > 0) {
                            /* Restarting a service does not bubble up a permissions error, but rather returns a cannot open error message which could be misleading. */
                            if (psErrorData.IndexOf("cannot be stopped due to the following error: Cannot open") > -1)
                            {
                                psErrorData = "[ERROR] Access may be denied to perform the requested action due to lack of permissions.  Additional error information has been included: " + psErrorData;
                            }
                            else
                            {
                                psErrorData = "[ERROR](Remote Error) " + psErrorData;
                            }
                        }
                        /* Gather response data */
                        foreach (PSObject obj in results)
                        {
                            psResult.AppendLine(obj.ToString());
                        }

                        powershell.Dispose();
                        runspace.Dispose();
                    }
                    catch (Exception ex)
                    {
                        if (ex.Message.Length > 0)
                        {
                            psErrorData = ex.Message.ToString();
                        }
                    }
                }
                else
                {
                    psErrorData = tmpComputerName;
                }
            }
            else
            {
                psErrorData = "[ERROR] Service name appears to be invalid.";
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }

        #region Helper Methods
        /// <summary>
        /// Determines if a string, representing an IP address, is valid
        /// </summary>
        /// <param name="address"></param>
        /// <returns></returns>
        private static bool IsValidIP(string address)
        {
            IPAddress ip;
            if (!IPAddress.TryParse(address, out ip)) return false;
            switch (ip.AddressFamily)
            {
                case AddressFamily.InterNetwork:
                    if (address.Length > 6 && address.Contains("."))
                    {
                        string[] s = address.Split('.');
                        if (s.Length == 4 && IsNumeric(s[0]) == true &&IsNumeric(s[1]) == true &&IsNumeric(s[2]) == true && IsNumeric(s[3]) == true)
                        {
                            return true;
                        }
                    }
                    break;
                case AddressFamily.InterNetworkV6:
                    if (address.Contains(":") && address.Length > 15)
                    {
                        return true;
                    }
                    break;
            }
            return false;
        }

        /// <summary>
        /// Ensure that the string submitted only contains digits.
        /// </summary>
        /// <param name="rawOctet"></param>
        /// <returns></returns>
        private static bool IsNumeric(string rawOctet)
        {
            bool result = false;
            Regex validDigits = new Regex("^[0-9]+$");
            if (rawOctet.Length > 0)
            {
                result = validDigits.IsMatch(rawOctet);
            }
            return result;
        }

        /// <summary>
        /// Ensure that the string submitted only contains specific characters for the name of a service.
        /// </summary>
        /// <param name="rawString"></param>
        /// <returns></returns>
        private static bool IsValidString(string rawString)
        {
            bool result = false;
            Regex validCharacters = new Regex("^[A-Za-z0-9()_.,/-]+$");
            if (rawString.Length > 0)
            {
                result = validCharacters.IsMatch(rawString);
            }
            return result;
        }

        /// <summary>
        /// Ensure that the string submitted representing a computername or IP address contain specific characters.
        /// </summary>
        /// <param name="rawString"></param>
        /// <returns></returns>
        private static bool IsValidComputerNameIPString(string rawString)
        {
            bool result = false;
            Regex validCharacters = new Regex("^[A-Za-z0-9()_.:-]+$");
            if (rawString.Length > 0)
            {
                result = validCharacters.IsMatch(rawString);
            }
            return result;
        }

        /// <summary>
        /// Validates local filepath and filename
        /// </summary>
        /// <param name="localFile"></param>
        /// <returns></returns>
        private static bool IsValidFilepath(string localFile)
        {
            bool result = true;
            var lastBackslash = localFile.LastIndexOf('\\');
            var dir = localFile.Substring(0, lastBackslash);
            var file = localFile.Substring(lastBackslash, localFile.Length - lastBackslash);
            if (file.IndexOf('\\') > -1)
            {
                file = file.Split('\\')[1];
            }
            /* Validate filepath */
            foreach (var p in Path.GetInvalidPathChars())
            {
                if (dir.IndexOf(p) > -1)
                {
                    result = false;
                }
            }
            /* Validate filename */
            foreach (var n in Path.GetInvalidFileNameChars())
            {
                if (file.IndexOf(n) > -1)
                {
                    result = false;
                }
            }
            return result;
        }

        /// <summary>
        /// Get the MAC address(s) of target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        private static StringBuilder GetMacAddress(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            tmpComputerName = computerName;

            /* Perform request */
            try
            {
                var runspace = RunspaceFactory.CreateRunspace();
                runspace.Open();
                var powershell = PowerShell.Create();
                powershell.Runspace = runspace;

                powershell.AddScript(String.Format("Get-WmiObject -ClassName Win32_NetworkAdapterConfiguration -Filter \"IPEnabled = 'True'\" -ComputerName \"{0}\" | Select-Object -Property MACAddress,Description", tmpComputerName));
                var results = powershell.Invoke();
                /* Catch remote errors such as no permissions */
                foreach (ErrorRecord current in powershell.Streams.Error)
                {
                    psErrorData = psErrorData + current.Exception.ToString();
                }
                if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                /* Gather response data */
                foreach (PSObject obj in results)
                {
                    psResult.AppendLine(obj.ToString());
                }
                /* Return error if present */
                if (psErrorData.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append(psErrorData);
                }
                powershell.Dispose();
                runspace.Dispose();
            }
            catch (Exception ex)
            {
                if (ex.Message.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append("[ERROR](Remote Error) " + ex.Message.ToString());
                }
            }
            return psResult;
        }

        /// <summary>
        /// Get BIOS information of target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        private static StringBuilder GetBiosInfo(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            tmpComputerName = computerName;

            /* Perform request */
            try
            {
                var runspace = RunspaceFactory.CreateRunspace();
                runspace.Open();
                var powershell = PowerShell.Create();
                powershell.Runspace = runspace;

                powershell.AddScript(String.Format("Get-WmiObject -ClassName Win32_Bios -ComputerName \"{0}\" | Select-Object Manufacturer,Name,SerialNumber,Version", tmpComputerName));
                var results = powershell.Invoke();
                /* Catch remote errors such as no permissions */
                foreach (ErrorRecord current in powershell.Streams.Error)
                {
                    psErrorData = psErrorData + current.Exception.ToString();
                }
                if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                /* Gather response data */
                foreach (PSObject obj in results)
                {
                    psResult.AppendLine(obj.ToString());
                }
                /* Return error if present */
                if (psErrorData.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append(psErrorData);
                }
                powershell.Dispose();
                runspace.Dispose();
            }
            catch (Exception ex)
            {
                if (ex.Message.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append("[ERROR](Remote Error) " + ex.Message.ToString());
                }
            }
            return psResult;
        }

        /// <summary>
        /// Get hardware information of target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        private static StringBuilder GetHardwareInfo(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            tmpComputerName = computerName;

            /* Perform request */
            try
            {
                var runspace = RunspaceFactory.CreateRunspace();
                runspace.Open();
                var powershell = PowerShell.Create();
                powershell.Runspace = runspace;

                powershell.AddScript(String.Format("Get-WmiObject -ClassName Win32_ComputerSystem -ComputerName \"{0}\"", tmpComputerName) +" | Select-Object Domain,Manufacturer,Model,Name,PrimaryOwnerName,@{'Name'='TotalPhysicalMemory (GB)'; 'Expression'={[string]::Format('{0:N0}',[math]::truncate($_.totalphysicalmemory / 1GB))}}");
                var results = powershell.Invoke();
                /* Catch remote errors such as no permissions */
                foreach (ErrorRecord current in powershell.Streams.Error)
                {
                    psErrorData = psErrorData + current.Exception.ToString();
                }
                if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                /* Gather response data */
                foreach (PSObject obj in results)
                {
                    psResult.AppendLine(obj.ToString());
                }
                /* Return error if present */
                if (psErrorData.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append(psErrorData);
                }
                powershell.Dispose();
                runspace.Dispose();
            }
            catch (Exception ex)
            {
                if (ex.Message.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append("[ERROR](Remote Error) " + ex.Message.ToString());
                }
            }
            return psResult;
        }

        /// <summary>
        /// Get processor information of target machine.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        private static StringBuilder GetProcessorInfo(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            tmpComputerName = computerName;

            /* Perform request */
            try
            {
                var runspace = RunspaceFactory.CreateRunspace();
                runspace.Open();
                var powershell = PowerShell.Create();
                powershell.Runspace = runspace;

                powershell.AddScript(String.Format("Get-WmiObject Win32_Processor -ComputerName \"{0}\" | Select-Object Caption,DeviceID,Manufacturer,MaxClockSpeed,Name,SocketDesignation", tmpComputerName));
                var results = powershell.Invoke();
                /* Catch remote errors such as no permissions */
                foreach (ErrorRecord current in powershell.Streams.Error)
                {
                    psErrorData = psErrorData + current.Exception.ToString();
                }
                if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                /* Gather response data */
                foreach (PSObject obj in results)
                {
                    psResult.AppendLine(obj.ToString());
                }
                /* Return error if present */
                if (psErrorData.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append(psErrorData);
                }
                powershell.Dispose();
                runspace.Dispose();
            }
            catch (Exception ex)
            {
                if (ex.Message.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append("[ERROR](Remote Error) " + ex.Message.ToString());
                }
            }
            return psResult;
        }

        /// <summary>
        /// Get OS information of target computer.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        private static StringBuilder GetOSInfo(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            tmpComputerName = computerName;

            /* Perform request */
            try
            {
                var runspace = RunspaceFactory.CreateRunspace();
                runspace.Open();
                var powershell = PowerShell.Create();
                powershell.Runspace = runspace;

                powershell.AddScript(String.Format("Get-CimInstance Win32_OperatingSystem -ComputerName \"{0}\" | Select-Object Caption,Version,BuildNumber,SerialNumber,Organization,RegisteredUser,SystemDirectory,WindowsDirectory", tmpComputerName));
                var results = powershell.Invoke();
                /* Catch remote errors such as no permissions */
                foreach (ErrorRecord current in powershell.Streams.Error)
                {
                    psErrorData = psErrorData + current.Exception.ToString();
                }
                if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                /* Gather response data */
                foreach (PSObject obj in results)
                {
                    psResult.AppendLine(obj.ToString());
                }
                /* Return error if present */
                if (psErrorData.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append(psErrorData);
                }
                powershell.Dispose();
                runspace.Dispose();
            }
            catch (Exception ex)
            {
                if (ex.Message.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append("[ERROR](Remote Error) " + ex.Message.ToString());
                }
            }
            return psResult;
        }

        /// <summary>
        /// Get the OU of the target computer.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        private static StringBuilder GetOUInfo(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;
            tmpComputerName = computerName;

            /* Perform request */
            try
            {
                var runspace = RunspaceFactory.CreateRunspace();
                runspace.Open();
                var powershell = PowerShell.Create();
                powershell.Runspace = runspace;

                powershell.AddScript(String.Format("Get-WmiObject -ClassName Win32_NetworkAdapterConfiguration -Filter \"IPEnabled = 'True'\" -ComputerName \"{0}\" | Select-Object -Property MACAddress,Description", tmpComputerName));
                var results = powershell.Invoke();
                /* Catch remote errors such as no permissions */
                foreach (ErrorRecord current in powershell.Streams.Error)
                {
                    psErrorData = psErrorData + current.Exception.ToString();
                }
                if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                /* Gather response data */
                foreach (PSObject obj in results)
                {
                    psResult.AppendLine(obj.ToString());
                }
                /* Return error if present */
                if (psErrorData.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append(psErrorData);
                }
                powershell.Dispose();
                runspace.Dispose();
            }
            catch (Exception ex)
            {
                if (ex.Message.Length > 0)
                {
                    psResult.Clear();
                    psResult.Append("[ERROR](Remote Error) " + ex.Message.ToString());
                }
            }
            return psResult;
        }

        /// <summary>
        /// Get the PowerShell version of the target computer.
        /// </summary>
        /// <param name="computerName"></param>
        /// <returns></returns>
        public static List<KeyValuePair<string, string>> PSGetVersion(string computerName = "")
        {
            /* PS Related Data */
            StringBuilder psResult = new StringBuilder();
            string psErrorData = String.Empty;
            string tmpComputerName = String.Empty;

            /* User computer name or resolve to computer name */
            tmpComputerName = PSGetComputerName(computerName);
            /* Do not continue if the computer name could not be resolved */
            if (tmpComputerName.IndexOf("[ERROR]", StringComparison.OrdinalIgnoreCase) == -1)
            {
                try
                {
                    var runspace = RunspaceFactory.CreateRunspace();
                    runspace.Open();
                    var powershell = PowerShell.Create();
                    powershell.Runspace = runspace;

                    if (tmpComputerName == "$env:ComputerName")
                    {
                        powershell.AddScript("Get-Host | Select-Object Version");
                    }
                    else
                    {
                        powershell.AddScript(String.Format("Invoke-Command -ComputerName \"{0}\" -ScriptBlock", tmpComputerName) + " { Get-Host | Select-Object Version }");
                    }
                    var results = powershell.Invoke();
                    /* Catch remote errors such as no permissions */
                    foreach (ErrorRecord current in powershell.Streams.Error)
                    {
                        psErrorData = psErrorData + current.Exception.ToString();
                    }
                    if (psErrorData.Length > 0) { psErrorData = "[ERROR](Remote Error) " + psErrorData; }
                    /* Gather response data */
                    foreach (PSObject obj in results)
                    {
                        psResult.Append(obj.ToString());
                    }

                    powershell.Dispose();
                    runspace.Dispose();
                }
                catch (Exception ex)
                {
                    if (ex.Message.Length > 0)
                    {
                        psErrorData = ex.Message.ToString();
                    }
                }
            }
            else
            {
                psErrorData = tmpComputerName;
            }

            /* Build Response */
            var responseList = new List<KeyValuePair<string, string>>()
            {
                new KeyValuePair<string, string>("psdata", psResult.ToString()),
                new KeyValuePair<string, string>("error", psErrorData)
            };

            return responseList;
        }
        #endregion
    }
}
